/*
 * Stream demultiplexer, line oriented, flock()s files before writing
 */
#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <signal.h>

#if STDC_HEADERS
# include <string.h>
#else
# if !HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# if !HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#if HAVE_UNISTD_H
# if HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
# include <unistd.h>
#endif

#if HAVE_SYSLOG_H
# include <syslog.h>
#endif

#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include <ctype.h>

/* #define DEBUG */

#include "demux_config.h"
#include "demux.h"

file *files;

void close_files(file *files)
{
	int i;

	for (i = 0; i < FILE_TABLE_SIZE; i++)
		if (files[i].handle)
			fclose(files[i].handle);

}

void cleanup(int sig)
{
	closelog();
	close_files(files);
	exit(0);
}

void init(char *p)
{
	openlog(p, LOG_PID, DEMUX_LEVEL);
	signal(SIGINT, cleanup);
	signal(SIGTERM, cleanup);
}

void error(char *msg)
{
	syslog(DEMUX_LEVEL, "%s", msg);
}

	/* valid tags contain alpha numerics or '-', returns: 0 - fail; 1 - OK */
int validate_tag(char *tag)
{
	int taglen;
	char *t;

	t = tag;
	taglen = strlen(t);
		/* no null tags allowed */
	if (taglen <= 0) {
		error("empty tag");
		return 0;
	}
	if (taglen > MAX_TAG_SIZE) {
		syslog(DEMUX_LEVEL, "tag %s too long", tag);
	}
	
	for (; *t; t++) {
		int ok;
		ok = isalnum(*t) || (*t == '-');
		if (!ok) {
			syslog(DEMUX_LEVEL, "invalid tag in %s (%c)", tag, *t);
			return 0;
		}
	}
	return 1;
}

u_int hash(char *name)
{
	int acc;
	for (acc = 0; *name; name++)
		acc += *name;
	
	return acc % FILE_TABLE_MOD;
}

int next_free_slot(file *files, int pos)
{
	int start;
	start = pos;
	do {
		if (!files[pos].name)
			return pos;
		pos = (pos + 1) % FILE_TABLE_MOD;
	}
	while (pos != start);
	return -1;
}

int find_free_slot(file *files, char *name)
{
	return next_free_slot(files, hash(name));
}

int find_file(file *files, char *name)
{
	int start, pos;
	pos = start = hash(name);
	do {
		if (files[pos].name) {
			if (strcmp(name, files[pos].name) == 0) {
				return pos;
			}
		}
		else
			return -1;
		pos = (pos + 1) % FILE_TABLE_MOD;
	} while (pos != start);
	
	return -1;
}

int add_file(file *files, char *name)
{
	int pos;
	
	if (!validate_tag(name)) {
		error("invalid filename supplied to add_file");
		return 0;
	}
	
	pos = find_free_slot(files, name);
	if (pos == -1) {
		error("file table is full");
		return 0;	/* table is full */
	}
	
	files[pos].handle = fopen(name, "a");
	if (!files[pos].handle) {
		error("cannot open output file");
		return 0;
	}
	files[pos].name = strdup(name);
	if (!files[pos].name) {
		error("cannot allocate memory for filename");
		return 0;
	}
	return 1;
}

file *setup_output_files(char *config)
{
	FILE *config_file;
	file *file_table;
	int file_table_byte_size;
	char line[MAX_LINE_SIZE];
	
	config_file = fopen(config, "r");
	if (!config_file) {
		error("cannot open configuration file");
		return NULL;
	}
	
	file_table_byte_size = sizeof(file) * FILE_TABLE_SIZE;
	file_table = (file *)malloc(file_table_byte_size);
	if (!file_table) {
		error("cannot allocate file table");
		return NULL;
	}
		/* initialize file list */
	memset(file_table, 0, file_table_byte_size);
	while (fgets(line, MAX_LINE_SIZE, config_file)) {
		int len;
		len = strlen(line);
		if (line[len - 1] == '\n')
			line[len - 1] = '\0';
		add_file(file_table, line);
	}
	
	fclose(config_file);
	return file_table;
}


void write_to_file(file *files, char *tag, char *msg)
{
	int pos, out, len;
	
	pos = find_file(files, tag);
	
	if (pos == -1 || !files[pos].handle) {
		syslog(DEMUX_LEVEL, "unknown tag: %s", tag);
		return;
	}
	
	len = strlen(msg);
	out = fileno(files[pos].handle);
	flock(out, LOCK_EX);
	
        if (fprintf(files[pos].handle, "%s", msg) != len)
		syslog(DEMUX_LEVEL, "cannot write to output file %s", files[pos].name);

	fflush(files[pos].handle);

        flock(out, LOCK_UN);

	return;
}


void demux_stream(FILE *in, file *files)
{
	char	tag[MAX_TAG_SIZE+1];
	char	incoming[MAX_LINE_SIZE];
	int	msg_start = 1;

	memset(tag, 0, MAX_TAG_SIZE+1);

	while (fgets(incoming, MAX_LINE_SIZE, stdin)) {
		char *msg = NULL;
		
		if (msg_start) {
			msg = strchr(incoming, ' ');
			msg_start = 0;
		}

		if (msg) {
			*msg = '\0';
			msg++;
			strncpy(tag, incoming, MAX_TAG_SIZE);
		}
		else {
			msg = incoming;
		}

			/* no message gets written without a valid tag */
		if (validate_tag(tag)) {
				/* no empty messages get written */
			if (strlen(msg) > 0) {
				write_to_file(files, tag, msg);
			}
		}

		if (msg[strlen(msg) - 1] == '\n') {
			tag[0] = '\0';
			msg_start = 1;
		}
	} 
}

/* options:
 * -d	specifies the working directory
 *
 */

int process_options(int argc, char **argv)
{
	int option;
	
	while ((option = getopt(argc, argv, "w:")) != -1) {
		switch (option) {
			case 'w':
				if (chdir(optarg) == -1) {
					error("cannot chdir\n");
				}
				break;
			default:
				error("unrecognized option");
				return;
		}
	}
	return optind;
}

void usage(char *p)
{
	fprintf(stderr, "usage: %s [-w working_dir] file-list\n", p);
	exit(1);
}

int main(int argc, char **argv)
{
	int last_arg;
	
	if (argc < 2) {
		usage(argv[0]);
	}
	
	init(argv[0]);
	last_arg = process_options(argc, argv);
	if ((argc - last_arg) != 1) {
		usage(argv[0]);
	}
		/* read configuration etc. */
	files = setup_output_files(argv[last_arg]);	
	if (!files) exit(1);
	demux_stream(stdin, files);
	
	cleanup(0);
}
