/*
 * stale.c
 *
 * Monitor files statefully
 * Copyright 2002, Jason Royes
 * 
 */
 
#include <stdio.h>

#if HAVE_CONFIG_H
# include "config.h"
#endif

	/* standard C headers */
#if STDC_HEADERS
# include <string.h>
#else
# if !HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# if !HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#if HAVE_UNISTD_H
# if HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
# include <unistd.h>
#endif

#if HAVE_LIBGEN_H
# include <libgen.h>
#endif

#include <sys/stat.h>
#include <signal.h>

#include "stale.h"

int	n_files;	/* numer of files to monitor */
struct	lf *lfs;	/* collective status */
char 	hostname[1024];	/* store hostname */

/*
 * -f	Don't bail on EOF
 * -r	Don't bail when a file 'does not exist'
 * -s	Tag each message with a syslog facility/priority
 * -i	Set working directory
 */

void usage(char *p)
{
	fprintf(stderr, "stale-error %s: usage: %s [-w working_dir] [-frt] log1 ... logN [tag1 ... tagN]\n", 
			hostname, p);
	bail(E_USAGE);
}

void clean_up(void)
{
	/* close all open file descriptors */
	int lfi;
	
	for (lfi = 0; lfi < n_files; lfi++) {
		if (lfs[lfi].f)
			fclose(lfs[lfi].f);
	}
}

void sig_handler(int s, siginfo_t *si, void *bleh)
{
	clean_up();
	exit(0);
}

	/* bail with message */
void bail(int exit_status)
{
	clean_up();
	exit(exit_status);
}

/* Restore state from a file */

int restore(char *filename, struct lf *log)
{
	FILE *f;
	int ok = 1;
	
	f = fopen(filename, "r");
	if (!f) return 0;
	if (fscanf(f, "%10u %10u", &(log->i), &(log->s)) < 2) { ok = 0; }
	fclose(f); 
	
	return ok;
}

/* Save state to file */

int save(char *filename, struct lf *log)
{
	FILE *f;
	int ok = 1;
	
	f = fopen(filename, "w");
	if (!f) {
		fprintf(stderr, "stale-error %s: cannot open %s for writing\n", hostname, filename);
		return 0;
	}
	if (fprintf(f, "%u %u\n", log->i, log->s) < 0) { ok = 0; }
	fclose(f);
	
	return ok;
}

/* Read a file up to EOF */

int bf(FILE *f, char *tag)
{
	int br = 0, len = 0;
	int line_start = 1;
	char line[BB_SIZE];

	while (fgets(line, BB_SIZE, f)) {
		len = strlen(line);
		if (tag && line_start) {
			printf("%s ", tag);
			line_start = 0;
		}
		printf("%s", line);
		if (tag && line[len - 1] == '\n') {
			line_start = 1;
		}
		br += len;
	}

	if (br)	fflush(stdout);
	return br;
}

void setup_signals()
{
	struct	sigaction ba;
		/* setup signal handler for SIGINT & TERM*/
	memset(&ba, 0, sizeof(struct sigaction));
	ba.sa_sigaction = sig_handler;
	ba.sa_flags = SA_SIGINFO;
	sigaction(SIGINT, &ba, NULL);
	sigaction(SIGTERM, &ba, NULL);
}

char *compute_state_file_name(char *original_name)
{
	char *state_file_basename;
	char *new_name;
	
	state_file_basename = basename(original_name);
		/* basename + extension + \0 */
	new_name = (char *)malloc(strlen(state_file_basename) + 
				strlen(STATE_EXT) + 1);
	if (!new_name) return NULL;
	new_name[0] = '\0';
	
	strcat(new_name, state_file_basename);
	strcat(new_name, STATE_EXT);
	return new_name;
}

#define LOG	argv[lfi + lf_start]

int main(int argc, char **argv) 
{
	int 	lfi; 	/* log file index */
	int	br;	/* bytes read */
	int	tags_start;	/* tags start at this arg position */
	int	lf_start;	/* log files start at this arg position */
	int	loop_forever;
	int	retry_files;
	int 	option;
	int	mode;	/* tail or streamlog mode */
	int	aux;
	char 	*state_dir;
	char 	**state_file_names;
	struct	stat s;
		
		/* gethostname should prolly be replaced with uname(), it seems to be blessed by POSIX */
#if HAVE_GETHOSTNAME
	gethostname(hostname, 1023);
	hostname[1023] = '\0';
#else
	strcpy(hostname, "noname");
#endif

	if (argc < 2)
		usage(argv[0]);

		/* initialize options to defaults */
	loop_forever = 0;
	retry_files = 0;
	mode = TAIL_MODE;
	state_dir = STATUS_DIR;
	
		/* process options */
	while ((option = getopt(argc, argv, "frtw:")) != -1) {
		switch (option) {
			case 'f':
				loop_forever = 1;
				break;
			case 'r':
				retry_files = 1;
				break;
			case 't':
				mode = TAG_MODE;
				break;
			case 'w':
				state_dir = optarg;
				break;
			default:
				usage(argv[0]);
		}
	}

	if (chdir(state_dir) == -1) {
		fprintf(stderr, "stale-error %s: chdir to %s failed\n", hostname, state_dir);
		exit(E_CHDIR);
	}
	
		/* log files start at first non-option */
	lf_start = optind;
	if (mode == TAG_MODE) {
			/* make sure even number of files+tags were passed in */
		if (argc <= lf_start || (argc - lf_start) % 2 != 0)
			usage(argv[0]);
			/* calculate number of files to monitor */
		n_files = (argc - lf_start) / 2;
		tags_start = n_files + lf_start;
	}
	else {
			/* ensure non-empty set of log files specified */
		if (argc - lf_start < 1)
			usage(argv[0]);
		n_files = argc - lf_start;
	}
	
		/* install signal handlers */
	setup_signals();
	
	lfs = (struct lf *)malloc((sizeof(struct lf) * n_files));
	if (!lfs) {
		fprintf(stderr, "stale-error %s: cannot allocate status memory\n", hostname);
		bail(E_NO_MEMORY);
	}
	state_file_names = (char **)malloc(sizeof(char *) * n_files);
	if (!state_file_names) {
		fprintf(stderr, "stale-error %s: cannot allocate status filename memory", hostname);
		bail(E_NO_MEMORY);
	}
		/* perform final initializations */
	memset(lfs, 0, sizeof(struct lf) * n_files);
	br = 0;
	
		/* open each file for reading and restore its status (if any) */
	for (lfi = 0; lfi < n_files; lfi++) {
		
		do {
			if (!(lfs[lfi].f = fopen(LOG, "r"))) {
				fprintf(stderr, "stale-error %s: cannot open file %s for reading\n", hostname, LOG);
				if (!retry_files)
					bail(E_OPEN_FILE);
				sleep(DELAY);
			}
		} while (retry_files && !lfs[lfi].f);
		
			/* compute state file name */
		state_file_names[lfi] = compute_state_file_name(LOG);
		if (!state_file_names[lfi]) {
			fprintf(stderr, "stale-error %s: cannot allocate memory for state filename", hostname);
			bail(E_NO_MEMORY);
		}
		
		if (!restore(state_file_names[lfi], &lfs[lfi])) {
				/* if no state was restored, get current */
			do {
				if (aux = stat(LOG, &s)) {
					fprintf(stderr, "stale-error %s: cannot stat %s\n", hostname, LOG);
					if (!retry_files)
						bail(E_STAT_FILE);
					sleep(DELAY);
				}
			} while (aux && retry_files);
			
			lfs[lfi].s = 0;
			lfs[lfi].i = s.st_ino;
		}
		else {	/* sync file position, SHOULD THIS BAIL ? */
			if (fseek(lfs[lfi].f, lfs[lfi].s, SEEK_SET)) {
				fprintf(stderr, "stale-error %s: cannot seek to restored position in %s\n", 
						hostname, LOG);
				bail(E_FSEEK);
			}
		}
	}
	
	do {
		for (lfi = 0; lfi < n_files; lfi++) {
			int truncated = 0;
			int changed = 0;
			
			if (stat(LOG, &s) == -1) {
				fprintf(stderr, "stale-error %s: cannot stat file: %s\n", hostname, LOG);
				if (!retry_files)
					bail(E_STAT_FILE);
			}
			else {	
				if (lfs[lfi].i != s.st_ino || ((lfs[lfi].s > s.st_size) && (truncated = 1))) {
					if (!(freopen(LOG, "r", lfs[lfi].f))) {
						fprintf(stderr, "stale-error %s: cannot re-open file %s for reading\n", hostname, LOG);
						bail(E_OPEN_FILE);
					}
					if (truncated)
						fprintf(stderr, "stale-error %s: file %s truncated\n", hostname, LOG);
#if DEBUG
					else
						fprintf(stderr, "stale-error %s: file %s inode changed\n", hostname, LOG);
#endif
					lfs[lfi].s = 0;
					lfs[lfi].i = s.st_ino;
					changed = 1;
				}
				if (lfs[lfi].s < s.st_size) {
					br = bf(lfs[lfi].f, (mode == TAG_MODE) ? argv[tags_start + lfi] : NULL);
					lfs[lfi].s += br;
					changed = 1;
				}
				if (changed && !save(state_file_names[lfi], &lfs[lfi])) {
					fprintf(stderr, "stale-error %s: error saving state\n", hostname);
					if (!loop_forever)
						bail(E_CANNOT_SAVE_STATE);
					changed = 0;
				}
			}
		}
		
		if (loop_forever)
			sleep(DELAY);
	} while (loop_forever);
	exit(0);
}
