#if HAVE_CONFIG_H
# include "config.h"
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include <openssl/ssl.h>
#include <openssl/err.h>

#include "sslhelp.h"
#include "tun.h"

	/* call tunnel_init(t) first */
int tunnel_up(tunnel *t)
{
//unsigned long ret;
	int ret;
	int i;

	if (!t || !t->t) {
		qerr("bad tunnel passed into tunnel_up");
		return E_TUN_BADTNL;
	}
	if (t->state & UP) {
		return 0;
	}
	
	i = time(NULL);
		/* avoid flooding server with connection requests */
	if ((i - t->last_connect) < CONNECT_TO) {
		return 0;
	}
	
#ifdef DEBUG
	qerr("connecting tunnel");
#endif
	t->last_connect = i;
	ret = create_socket(t->host, t->port, &t->s);
	if (ret) {
#ifdef DEBUG
		qerr("create_socket");
#endif
		return ret;
	}
	if (SSL_set_fd(t->t, t->s) == 0) {
#ifdef DEBUG
		qerr("SSL_set_fd");
#endif
		return E_TUN_SETFD;
	}
		/* do the SSL handshake */
	ret = SSL_connect(t->t);
	if (ret != 1) {
		syslog(TUN_PRI, "SSL handshake failed (%s)", 
				ERR_error_string(ERR_get_error(), NULL));
		return E_TUN_CONNECT;
	}
		/* verify authentication result */
#if VERIFY_CERT
	ret = SSL_get_verify_result(t->t);
	if (ret != X509_V_OK) {
		syslog(TUN_PRI, "could not verify server certificate (%s)", 
				ERR_error_string(ERR_get_error(), NULL));
		return E_TUN_AUTH;
	}
#endif
	t->state = UP;
	queue_close(&t->q);
	
	if (ret = tunnel_flush(t)) {
#ifdef DEBUG
		qerr("non-zero exit from tunnel_flush");
#endif
		return ret;
	}
	return 0;
}

int tunnel_down(tunnel *t)
{
	if (!t || !t->t)
		return E_TUN_BADTNL;
	if (t->state & DOWN)
		return 0;
	if (t->s != -1) {
		SSL_shutdown(t->t);
		close(t->s);
	}
	t->state = DOWN;
	queue_open(&t->q, O_WRONLY|O_CREAT|O_APPEND);
	return 0;
}

/* Returns 0 on error
 */
int tunnel_init(SSL_CTX *ctx, tunnel *t, char *host, uint16_t port)
{
	if (!ctx || !host)
		return 0;
	t->t = SSL_new(ctx);
	if (!t->t)
		return 0;
	t->state = DOWN;
	t->host = host;
	t->port = port;
	t->s = -1;
	
	t->q.q_fd = -1;
	t->q.q_state_name = Q_STATE_NAME;
	t->q.q_name = Q_NAME;
	queue_get_size(&t->q);
	queue_open(&t->q, O_WRONLY|O_CREAT|O_APPEND);
	
	t->last_write = 0;
	t->last_connect = 0;
	return 1;
}

int tunnel_write(tunnel *t, char *data, int len)
{
	tunnel_up(t);
	if (t->state & UP) {
		if (SSL_get_shutdown(t->t) == SSL_RECEIVED_SHUTDOWN) {
			tunnel_down(t);
			return E_TUN_BADTNL;
		}
		if (SSL_write(t->t, data, len) != len) {
				/* there's something wrong with the tunnel */
			if (t->state & FLUSHING)
				return E_TUN_STOPFLUSH;
			tunnel_down(t);
			tunnel_write(t, data, len);
			return E_TUN_WRITE;
		}
		else {
			t->last_write = time(NULL);
		}
		return 0;
	}
	else if (t->state & DOWN) {
		if (t->state & FLUSHING)
			return E_TUN_STOPFLUSH;
		return queue_write(&t->q, data, len);
	}
	/* unreachable */
	exit(11);
}

	/* Flush a tunnel's queue */
int tunnel_flush(tunnel *t)
{
	int q_state_fd, ret;
	unsigned int len, q_seek_pos;
	char buf[FLUSH_BUF_SIZE];
	
	if (t->state != UP)
		return 0;
	
	q_seek_pos = 0;
	if (t->q.q_fd != -1)
		return E_TUN_QISOPEN;
	q_state_fd = open(t->q.q_state_name, O_RDONLY);
	if (q_state_fd != -1) {
		len = read(q_state_fd, &q_seek_pos, sizeof(q_seek_pos));
		close(q_state_fd);
		if (len != sizeof(q_seek_pos))
			return E_TUN_QCORRUPT;
	}
	
	if (ret = queue_open(&t->q, O_RDONLY)) {
		return E_TUN_NOQFILE;
	}
	
		/* restore state if found and if queue file opens successfully */
	q_state_fd = open(t->q.q_state_name, O_WRONLY|O_CREAT, 0600);
	if (q_state_fd == -1)
		return E_TUN_QSTATE;
	
	if (q_seek_pos && (lseek(t->q.q_fd, q_seek_pos, SEEK_SET) != q_seek_pos))
		return E_TUN_BADSEEK;

	t->state |= FLUSHING;

#ifdef DEBUG
	qerr("flushing queue");
#endif

	while ((len = read(t->q.q_fd, buf, FLUSH_BUF_SIZE)) > 0) {
		ret = tunnel_write(t, buf, len);
		if (ret == E_TUN_STOPFLUSH) {
			/* lost tunnel while flushing, close: q, q-state and bail */
			qerr("lost tunnel while flushing");
			queue_close(&t->q);
			close(q_state_fd);
			return E_TUN_STOPFLUSH;
		}
		if (write(q_state_fd, &q_seek_pos, sizeof(q_seek_pos)) != sizeof(q_seek_pos)) {
			/* could not save state, close: queue, queue-state and bail */
			qerr("could not save state");
			queue_close(&t->q);
			close(q_state_fd);
			tunnel_down(t);
			return E_TUN_WRSTATE;
		}
		fsync(q_state_fd);		/* flush to disk */
		q_seek_pos += len;
		lseek(q_state_fd, 0, SEEK_SET);	/* seek to first byte */
	}

#ifdef DEBUG
	qerr("flush complete");
#endif
	close(q_state_fd);
	queue_close(&t->q);
	unlink(t->q.q_name);
	unlink(t->q.q_state_name);
	t->state = UP;
	return 0;
}

int queue_close(queue *q)
{
	if (q->q_fd == -1)
		return E_TUN_QCLOSED;
	close(q->q_fd);
	q->q_fd = -1;
	return 0;
}

int queue_get_size(queue *q)
{
	struct stat qs;
	
	if (stat(q->q_name, &qs) != 0) {
		q->size = 0;
		return 1;
	}
	
	q->size = qs.st_size;
	return 0;
}

int queue_open(queue *q, int mode)
{
		
	if (q->q_fd != -1)
		return E_TUN_QISOPEN;
	q->q_fd = open(q->q_name, mode, 0600);
	if (q->q_fd == -1)
		return E_TUN_NOQFILE;
	
	queue_get_size(q);
	return 0;
}

int queue_write(queue *q, char *data, int len)
{
	if (q->q_fd == -1)
		return E_TUN_QNOTOPEN;
	if (q->size + len > Q_SIZE_MAX)
		return E_TUN_QFULL;
	if (write(q->q_fd, data, len) != len) {
		queue_close(q);
		return E_TUN_QWRITE;
	}
	q->size += len;
	return 0;
}
